import enum
import gzip
import logging
import os
import re
import pysam

logger = logging.getLogger('utils')


class NGSFileType(enum.Enum):
    FASTQ = 1
    SAM = 2
    BAM = 3


class NGSFile:
    def __init__(self, filename):
        self.filename = filename
        self.basename = os.path.basename(self.filename)
        self.ext = ".".join(self.basename.split(".")[1:])
        self.readmode = "r"
        self.gzipped = False

        if self.ext.endswith(".gz") or self.ext.endswith(
                ".bgz") or self.ext.endswith(".bam"):
            self.readmode = "rb"
            self.gzipped = True

        if self.ext.endswith("fastq") or \
           self.ext.endswith("fq") or \
           self.ext.endswith("fastq.gz") or \
           self.ext.endswith("fq.gz"):
            self.filetype = NGSFileType.FASTQ
            if self.gzipped:
                self.handle = gzip.open(self.filename, mode=self.readmode)
            else:
                self.handle = open(self.filename, mode=self.readmode)
        elif self.ext.endswith("sam"):
            self.filetype = NGSFileType.SAM
            self.handle = pysam.AlignmentFile(self.filename, self.readmode)
        elif self.ext.endswith("bam"):
            self.filetype = NGSFileType.BAM
            self.handle = pysam.AlignmentFile(self.filename, self.readmode)
        else:
            raise RuntimeError("Could not determine NGS file type: {}".format(
                self.filename))

        logger.debug("Opened NGS file '{}' as {}".format(
            self.filename, self.filetype))
        logger.debug("Gzipped: {}".format(self.gzipped))
        logger.debug("Readmode: {}".format(self.readmode))

    def __iter__(self):
        self.read_num = 0
        return self

    def __next__(self):
        if self.filetype == NGSFileType.FASTQ:
            query_name = self.handle.readline().strip()
            if not query_name or query_name == "":
                raise StopIteration()

            query = self.handle.readline().strip()
            plusline = self.handle.readline().strip()
            quality = self.handle.readline().strip()

            if self.gzipped:
                query_name = query_name.decode("utf-8")
                query = query.decode("utf-8")

            if query_name.startswith("@"):
                query_name = query_name[1:]

        elif self.filetype == NGSFileType.SAM or self.filetype == NGSFileType.BAM:
            read = next(self.handle)
            query_name = read.query_name
            query = read.query

        self.read_num += 1

        return {"query_name": query_name, "query": query}


class GFF:
    def __init__(self,
                 filename,
                 feature_type=None,
                 filters=None,
                 gene_blacklist=None):
        self.gene_blacklist = None
        if gene_blacklist:
            self.gene_blacklist = set([
                item.strip() for item in open(gene_blacklist, 'r').readlines()
            ])
        self._handle = gzip.open(filename, "r")
        self.feature_type = feature_type
        self.filters = filters
        self.entries = []
        self.attr_regexes = [r"(\S+)=(\S+)", r"(\S+) \"(\S+)\""]

        while True:
            line = self._handle.readline().decode("utf-8")
            if not line:
                break

            if line.startswith("#"):
                continue

            [
                seqname, source, feature, start, end, score, strand, frame,
                attribute
            ] = line.split("\t")

            if self.feature_type and feature != self.feature_type:
                continue

            entry_passes_gene_blacklist = True
            if self.gene_blacklist:
                for bad_gene in self.gene_blacklist:
                    if bad_gene in attribute:
                        entry_passes_gene_blacklist = False
                        break

            if not entry_passes_gene_blacklist:
                continue

            result = {
                "seqname": seqname,
                "source": source,
                "feature": feature,
                "start": int(start),
                "end": int(end),
                "score": score,
                "strand": strand,
                "frame": frame
            }

            for attr_raw in [s.strip() for s in attribute.split(";")]:
                if not attr_raw or attr_raw == "":
                    continue

                for regex in self.attr_regexes:
                    match = re.match(regex, attr_raw)
                    if match:
                        key, value = match.group(1), match.group(2)
                        result["attr_" + key] = value.strip()

            entry_passes_filters = True
            if entry_passes_filters and self.filters:
                for (k, v) in self.filters.items():
                    if k not in result or result[k] != v:
                        entry_passes_filters = False
                        break

            if entry_passes_filters:
                self.entries.append(result)

    def filter(self, func):
        self.entries = filter(func, self.entries)

    def __iter__(self):
        self.i = 0
        return self

    def __next__(self):
        if self.i < len(self.entries):
            self.i += 1
            return self.entries[self.i - 1]
        else:
            raise StopIteration



import csv
import itertools
import logging
import pysam
import random
import tabix
import sys

from collections import defaultdict

from ..utils import NGSFile, NGSFileType, GFF

logger = logging.getLogger('strandedness')


def get_filtered_reads_from_region(samfile,
                                   gene,
                                   min_quality=30,
                                   apply_filters=True):
    for read in samfile.fetch(gene['seqname'], gene["start"], gene["end"]):
        if apply_filters and (read.is_qcfail or read.is_duplicate
                              or read.is_secondary or read.is_unmapped
                              or read.mapq < min_quality):
            continue
        yield read


def disqualify_gene(gene, gff_tabix, only_consider_protein_genes=True):
    # potentially only consider protein coding genes.
    if only_consider_protein_genes:
        if not "attr_gene_type" in gene:
            return True
        if not "protein" in gene['attr_gene_type']:
            return True

    # if there are overlapping features on the positive and negative strand
    # ignore this gene.
    try:
        hits = gff_tabix.query(gene['seqname'], gene['start'], gene['end'])
    except:
        raise RuntimeError("Tabix querying failed. Does the index exist?")

    has_positive_gene = None
    has_negative_gene = None

    for hit in hits:
        # must be a gene
        if not hit[2] == "gene":
            continue

        if hit[6] == "+":
            has_positive_gene = hit
        elif hit[6] == "-":
            has_negative_gene = hit

        if has_positive_gene and has_negative_gene:
            break

    if has_positive_gene and has_negative_gene:
        return True

    return False


def get_reads_rg(read, default="unknown"):
    for (k, v) in read.tags:
        if k == "RG":
            return v

    return default


def get_predicted_strandedness(forward_evidence_pct, reverse_evidence_pct):
    predicted = "Inconclusive"
    if 0.4 <= forward_evidence_pct and forward_evidence_pct <= 0.6:
        predicted = "Unstranded"
    if 0.8 <= forward_evidence_pct:
        predicted = "Stranded-Forward"
    elif 0.8 <= reverse_evidence_pct:
        predicted = "Stranded-Reverse"

    return predicted


def determine_strandedness(ngsfilepath, gff, gff_tabix, n_genes=100, min_mapq=30,
                           minimum_reads_per_gene=10, split_by_rg=False,
                           only_protein_coding_genes=True, max_iterations_per_try=1000):
    try:
        ngsfile = NGSFile(ngsfilepath)
    except:
        result = {
            "File": ngsfilepath,
            "TotalReads": "N/A",
            "ForwardPct": "N/A",
            "ReversePct": "N/A",
            "Predicted": "Error opening file."
        }

        if split_by_rg:
            result["ReadGroup"] = "N/A"

        return [result]

    if ngsfile.filetype != NGSFileType.BAM:
        raise RuntimeError(
            "Invalid file: {}. `strandedness` only supports aligned BAM files!"
            .format(ngsfilepath))
    samfile = ngsfile.handle

    n_tested_genes = 0
    n_reads_observed = 0
    gene_blacklist = set()

    read_groups = ["unknown"]
    if 'RG' in samfile.header:
        read_groups += [rg['ID'] for rg in samfile.header['RG']]

    overall_evidence = defaultdict(dict)

    for rg in read_groups:
        overall_evidence[rg] = defaultdict(int)

    logger.debug("Starting sampling...")
    total_iterations = 0
    while True:
        total_iterations += 1
        if total_iterations > max_iterations_per_try:
          logger.warn("Max iterations reached! Moving forward with prediction.")
          break

        if n_tested_genes >= n_genes:
            break

        gene = random.choice(gff.entries)

        if gene["attr_gene_id"] in gene_blacklist:
            continue

        if disqualify_gene(
                gene,
                gff_tabix,
                only_consider_protein_genes=only_protein_coding_genes):
            continue

        logging.debug("== Candidate Gene ==")
        logging.debug("  [*] Name: {}".format(gene['attr_gene_name']))
        logging.debug("  [*] Location: {}:{}-{} ({})".format(
            gene['seqname'], gene['start'], gene['end'], gene['strand']))
        logging.debug("  [*] Actions:")

        gene_blacklist.add(gene["attr_gene_id"])
        relevant_reads = get_filtered_reads_from_region(samfile,
                                                        gene,
                                                        min_quality=min_mapq)

        reads_in_gene = 0
        this_genes_evidence = defaultdict(dict)
        for rg in read_groups:
            this_genes_evidence[rg] = defaultdict(int)

        for read in relevant_reads:
            reads_in_gene += 1
            if read.is_paired:
                if read.is_read1:
                    read_id = "1"
                elif read.is_read2:
                    read_id = "2"
                else:
                    raise RuntimeError("Read is not read 1 or read 2?")
            else:
                # SE reads are equivalent to just assuming the read is read 1.
                read_id = "1"

            if not read.is_reverse:
                read_strand_id = "+"
            else:
                read_strand_id = "-"

            gene_strand_id = gene['strand']
            reads_observed_state = read_id + read_strand_id + gene_strand_id
            this_reads_rg = get_reads_rg(read)
            this_genes_evidence[this_reads_rg][reads_observed_state] += 1

        if reads_in_gene >= minimum_reads_per_gene:
            logging.debug("    - Sufficient read count ({} >= {})".format(
                reads_in_gene, minimum_reads_per_gene))
            logging.debug("    - {}".format(" ".join([
                "{}:{}".format(k, v) for k, v in this_genes_evidence.items()
            ])))

            for rg in this_genes_evidence.keys():
                for state in this_genes_evidence[rg].keys():
                    overall_evidence[rg][state] += this_genes_evidence[rg][
                        state]

            n_tested_genes += 1
            n_reads_observed += reads_in_gene
        else:
            logging.debug("    - Read count too low ({} < {})".format(
                reads_in_gene, minimum_reads_per_gene))

    if split_by_rg:
        results = []
        for rg in read_groups:
            evidence_stranded_forward = overall_evidence[rg][
                "1++"] + overall_evidence[rg]["1--"] + overall_evidence[rg][
                    "2+-"] + overall_evidence[rg]["2-+"]
            evidence_stranded_reverse = overall_evidence[rg][
                "1+-"] + overall_evidence[rg]["1-+"] + overall_evidence[rg][
                    "2++"] + overall_evidence[rg]["2--"]
            total_reads = evidence_stranded_forward + evidence_stranded_reverse
            if total_reads == 0 and rg == "unknown":
                continue

            forward_pct = 0 if total_reads <= 0 else round(
                evidence_stranded_forward / total_reads, 4)
            reverse_pct = 0 if total_reads <= 0 else round(
                evidence_stranded_reverse / total_reads, 4)
            predicted = get_predicted_strandedness(forward_pct, reverse_pct)

            results.append({
                "File": ngsfilepath,
                "ReadGroup": rg,
                "TotalReads": total_reads,
                "ForwardPct": forward_pct,
                "ReversePct": reverse_pct,
                "Predicted": predicted
            })
        return results
    else:
        evidence_stranded_forward = 0
        evidence_stranded_reverse = 0

        for rg in read_groups:
            evidence_stranded_forward += overall_evidence[rg][
                "1++"] + overall_evidence[rg]["1--"] + overall_evidence[rg][
                    "2+-"] + overall_evidence[rg]["2-+"]
            evidence_stranded_reverse += overall_evidence[rg][
                "1+-"] + overall_evidence[rg]["1-+"] + overall_evidence[rg][
                    "2++"] + overall_evidence[rg]["2--"]

        total_reads = evidence_stranded_forward + evidence_stranded_reverse
        forward_pct = 0 if total_reads <= 0 else round(evidence_stranded_forward / total_reads, 4)
        reverse_pct = 0 if total_reads <= 0 else round(evidence_stranded_reverse / total_reads, 4)
        predicted = get_predicted_strandedness(forward_pct, reverse_pct)

        return [{
            "File": ngsfilepath,
            "TotalReads": total_reads,
            "ForwardPct": forward_pct,
            "ReversePct": reverse_pct,
            "Predicted": predicted
        }]


def main(ngsfiles,
         gene_model_file,
         outfile=sys.stdout,
         delimiter="\t",
         n_genes=100,
         minimum_reads_per_gene=10,
         only_protein_coding_genes=True,
         min_mapq=30,
         split_by_rg=False,
         max_tries=3,
         max_iterations_per_try=1000):
    logger.info("Arguments:")
    logger.info("  - Gene model file: {}".format(gene_model_file))
    logger.info("  - Number of genes: {}".format(n_genes))
    logger.info(
        "  - Minimum reads per gene: {}".format(minimum_reads_per_gene))
    logger.info("  - Only consider protein coding genes: {}".format(
        only_protein_coding_genes))
    logger.info("  - Minimum MAPQ: {}".format(min_mapq))
    logger.info("  - Split by RG: {}".format(split_by_rg))

    if max_iterations_per_try < n_genes:
      logger.error("Max iteration per try cannot be less than number of genes to search!")
      sys.exit(1)

    logger.info("Reading gene model...")
    gff = GFF(gene_model_file, feature_type="gene")
    logger.info("  - {} features processed.".format(len(gff.entries)))
    gff_tabix = tabix.open(gene_model_file)
    logger.info("  - Tabix loaded for feature lookup.")

    fieldnames = ["TotalReads", "ForwardPct", "ReversePct", "Predicted"]
    if split_by_rg:
        fieldnames = ["ReadGroup"] + fieldnames
    fieldnames = ["File"] + fieldnames

    writer = csv.DictWriter(outfile,
                            fieldnames=fieldnames,
                            delimiter=delimiter)
    writer.writeheader()
    outfile.flush()

    for ngsfilepath in ngsfiles:
        tries_for_file = 0

        while True:
            tries_for_file += 1
            logger.info("Processing {}, try #{}...".format(ngsfilepath, tries_for_file))

            entries = determine_strandedness(
                ngsfilepath,
                gff,
                gff_tabix,
                n_genes=n_genes,
                min_mapq=min_mapq,
                minimum_reads_per_gene=minimum_reads_per_gene,
                split_by_rg=split_by_rg,
                only_protein_coding_genes=only_protein_coding_genes,
                max_iterations_per_try=max_iterations_per_try)

            entries_contains_inconclusive = False
            for entry in entries:
                if entry["Predicted"] == "Inconclusive":
                    entries_contains_inconclusive = True
                    break

            if entries_contains_inconclusive and tries_for_file < max_tries:
                continue

            for entry in entries:
                writer.writerow(entry)
                outfile.flush()

            break



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ngsderive', 'ngsderive.commands', 'ngsderive.readers']

package_data = \
{'': ['*']}

install_requires = \
['colorlog>=4.0.2,<5.0.0',
 'pysam>=0.15.3,<0.16.0',
 'pytabix>=0.0.2,<0.0.3',
 'rstr>=2.2.6,<3.0.0']

entry_points = \
{'console_scripts': ['ngsderive = ngsderive.__main__:run']}

setup_kwargs = {
    'name': 'ngsderive',
    'version': '1.0.2',
    'description': 'Forensic analysis tool useful in backwards computing information from next-generation sequencing data.',
    'long_description': '# ngsderive\n\n[![PyPI Version](https://img.shields.io/pypi/v/ngsderive.svg)](https://pypi.python.org/pypi/ngsderive/)\n[![PyPI Python Versions](https://img.shields.io/pypi/pyversions/ngsderive.svg)](https://pypi.python.org/pypi/ngsderive/)\n[![PyPI Project Status](https://img.shields.io/pypi/status/ngsderive.svg)](https://pypi.python.org/pypi/ngsderive/)\n\n\n[![CI\nstatus](https://github.com/claymcleod/ngsderive/workflows/CI/badge.svg)](https://github.com/claymcleod/ngsderive/actions)\n\n`ngsderive` is a forensic analysis tool useful in backwards computing information \nfrom next-generation sequencing data. Notably, results are provided as a \'best guess\' — the tool does \nnot claim 100% accuracy and results should be considered with that understanding.\n\nNote that this utility only implements commands which were not available at the \ntime of writing in common NGS utilities (e.g. [Picard](https://broadinstitute.github.io/picard/)).\n\n## Getting Started\n\nThese instructions will get you a copy of the project up and running on your\nlocal machine for development and testing purposes. See deployment for notes on\nhow to deploy the project on a live system.\n\n### Installing\n\nTo get started with `ngsderive`, you can install it using pip:\n\n```bash\npip install git+https://github.com/claymcleod/ngsderive.git\n```\n\n### Development\n\nTo get started on a development version of the code, just run the following:\n\n```bash\nconda create -n ngsderive-dev python=3.7 poetry -y\nconda activate ngsderive-dev\ngit clone git@github.com:claymcleod/ngsderive.git\ncd ngsderive\npoetry install\n```\n## Usage\n\n### Illumina machine type\n\nThe `instrument` subcommand will attempt to backward compute the machine that\ngenerated a NGS file using (1) the instrument id(s) and (2) the flowcell id(s). \n\n### Limitations\n\n* This command may not comprehensively detect the correct machines as there is\n  no published catalog of Illumina serial numbers. As we encounter more serial\n  numbers in practice, we update this code.\n\n### Read length calculation\n\nThe `readlen` subcommand can be used to compute the read length used during\nsequencing can be estimated (or reported as inconclusive). At the time of\nwriting, the algorithm used is roughly:\n\n1. Compute distribution of read lengths for the first `--n-samples` reads in a\n   file.\n2. Assuming read length in the file can only decrease from the actual read\n   length (from adapter trimming or similar), the putative maximum read length\n   is considered to be the highest detected read length.\n3. If the percentage of reads that are evidence for the putative maximum read\n   length makes up at least `--majority-vote-cutoff`% of the reads, the putative\n   read length is considered to be confirmed. If not, the consensus read length\n   will be return as -1 (could not determine).\n   1. For example, if 100bp is the maximum read length detected and 85% percent\n      of the reads support that claim, then we considered 100bp as the consensus\n      read length. If only 30% of the reads indicated 100bp, the tool cannot\n      report a consensus).\n\n### Strandedness inference\n\nThe `strandedness` subcommand can be used to determine strandedness of RNA-Seq\nsamples. Note that due to the need to (1) to examine the alignment status of a\nread [not included in FASTQ] and (2) fetch regions using random access [not\navailable with SAM files], only BAM files are currently supported for this\nsubcommand.\n\nStrandedness can estimated by observing the following characteristics of a\nparticular read:\n\n* Whether the read is read 1 or read 2 ("read ordinal").\n* Whether the read was aligned to the + or - strand ("read strand")\n* Given a gene model, whether a feature of interest (usually a gene) falls on\n  the + or - strand ("gene strand").\n\nA shorthand notation for the state of a read can be achieved by simply\nconcatenating the three characteristics above (e.g., `1+-` means that a read 1\nwas aligned to the positive strand and a gene was observed at the same location\non the negative strand).\n\nGiven the notation above, the following lookup table can be used to see whether\na read is evidence for forward-strandedness or reverse-strandedness:\n\n| Patterns                   | Evidence for strandedness type |\n| -------------------------- | ------------------------------ |\n| `1++`, `1--`, `2+-`, `2-+` | Forward                        |\n| `2++`, `2--`, `1+-`, `1-+` | Reverse                        |\n\nBy default, the strandedness check is designed to work with the\n[GENCODE][gencode-website] geneset. Either a GTF or GFF file can be used as the\ngene model — you can use the following one liner to prepare the latest geneset\nfor hg38.\n\n```bash\n\n# hg38\ncurl ftp://ftp.ebi.ac.uk/pub/databases/gencode/Gencode_human/release_32/gencode.v32.annotation.gff3.gz | gunzip -c | sort -k1,1 -k4,4n -k5,5n | bgzip > gencode.v32.annotation.gff3.gz\ntabix -p gff gencode.v32.annotation.gff3.gz\n```\n\nIf you would like to use the script on hg19, it takes a little more finesse (given the different formats of the attribute column between versions):\n\n```bash\n# hg19\ncurl ftp://ftp.ebi.ac.uk/pub/databases/gencode/Gencode_human/release_32/GRCh37_mapping/gencode.v32lift37.annotation.gtf.gz | gunzip -c | sort -k1,1 -k4,4n -k5,5n | python <(cat \n<<END    \nimport re \nimport sys\n\nfor line in [l.strip() for l in sys.stdin]:\n  if line.startswith("#"):\n    print(line)                     \n  else:                       \n    columns = line.split(\'\\t\')\n    if len(columns) != 9:                                                    \n      raise RuntimeError("Unexpected column number: {}".format(len(columns)))\n    \n    print(\'\\t\'.join(columns[0:8]), end="\\t")\n    \n    attrs_post = []\n    for attr in columns[8].split(";"):                        \n      groups = re.match(r"\\s?(\\S+) (\\S+)\\s?", attr)\n      if groups:             \n        key = groups.group(1)                    \n        value = groups.group(2).replace("\\"", "").replace(" ", ",")\n        attrs_post.append(key + "=" + value)\n\n    print(";".join(attrs_post))\nEND\n) | sed \'s/^chrM/chrMT/g\' | sed \'s/^chr//g\' | bgzip > gencode.v32lift37.annotation.gtf.gz\ntabix -p gff gencode.v32lift37.annotation.gtf.gz\n```\n\nAt the time of writing, the algorithm works roughly like this:\n\n1. The gene model is read in and only `gene` features are retained.\n2. For `--n-genes` times, a randomly sampled gene is selected from the gene\n   model. The gene must pass a quality check. Of particular interest,\n   1. The gene must not be an overlapping feature on the opposite strand which\n      would present ambiguous results.\n   2. *Optionally*, the gene must be a protein coding gene.\n   3. *Optionally*, the gene must have at least `--minimum-reads-per-gene`\n      minimum reads per gene.\n3. All of the reads from that region of the genome are extracted and put through\n   several quality filters including but not limited to:\n   1. The read must not be marked as QC-failed.\n   2. The read must not be marked as a duplicate.\n   3. The read must not be marked as secondary.\n   4. The read must not be unmapped.\n   5. *Optionally*, the read have a minimum MAPQ score.\n4. For all reads that pass the above filters, compute the evidence and tally\n   results.\n\nThis lookup table is used for the classification of strandedness based on the evidence:\n\n| Lookup                            | Value              |\n| --------------------------------- | ------------------ |\n| 40% <= `forward_reads_pct` <= 60% | `Unstranded`       |\n| 80% <= `forward_reads_pct`        | `Stranded-Forward` |\n| 80% <= `reverse_reads_pct`        | `Stranded-Reverse` |\n| Else                              | `Inconclusive`     |\n\nThe tool will repeat the strandedness test at most `--max-tries` times to try to find a\nnon-`Inconclusive`prediction.\n\n### Differences from popular methods\n\nThe most popular strandedness inference tool that the author is aware of is\nRSeQC\'s\n[infer_experiment.py](http://rseqc.sourceforge.net/#infer-experiment-py). The\nmain difference is that RSeQC starts at the beginning of the BAM file and takes\nthe first `n` reads that match its criteria. If the BAM is coordinate sorted,\nthis would mean its not uncommon to have all of the evidence at the beginning of\n`chr1`. Anecdotally, this method differs in that it is slightly slower than\n`infer_experiment.py` but is expected to be more robust to biases caused by\nwhich reads are sampled.\n\n\n### Limitations\n\n* Does not yet work with single-end data (simply because the author doesn\'t have\n  any on hand). The tool will throw an error if any unpaired reads are\n  discovered (let us know in the issues if you need this supported).\n* Though hundreds of Unstranded and Stranded-Reverse data has been tested and\n  verified, Stranded-Forward data has not been tested to work with this tool\n  (simply because the author doesn\'t have on hand). We do not anticipate any\n  issues since Stranded-Reverse is working well.\n\n## Running the tests\n\n```bash\n> py.test\n```\n\n## Contributing\n\nPlease read [CONTRIBUTING.md](CONTRIBUTING.md) for details on our code of\nconduct, and the process for submitting pull requests to us.\n\n\n## Versioning\n\nWe use [SemVer](http://semver.org/) for versioning. For the versions available, see the [tags on this repository](https://github.com/your/project/tags).\n\n## License\n\nThis project is licensed as follows:\n* All code related to the `instrument` subcommand is licensed under the [AGPL\n  v2.0][agpl-v2]. This is not due any strict requirement, but out of deference\n  to some [code][10x-inspiration] I drew inspiration from (and copied patterns\n  from), the decision was made to license this code consistently.\n* The rest of the project is licensed under the MIT License - see the\n  [LICENSE.md](LICENSE.md) file for details.\n\n[10x-inspiration]:\nhttps://github.com/10XGenomics/supernova/blob/master/tenkit/lib/python/tenkit/illumina_instrument.py\n[agpl-v2]: http://www.affero.org/agpl2.html\n[gencode-website]:\nhttps://www.gencodegenes.org/\n',
    'author': 'Clay McLeod',
    'author_email': 'Clay.McLeod@STJUDE.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/claymcleod/ngsderive',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3

import argparse
import csv
import logging
import sys

from ngsderive import utils
from ngsderive.commands import readlen, instrument, strandedness

logger = logging.getLogger('ngsderive')

def get_args():
  class SaneFormatter(argparse.RawTextHelpFormatter, argparse.ArgumentDefaultsHelpFormatter):
    pass

  parser = argparse.ArgumentParser(description="Derive useful information (or best guess) from next-generation sequencing files.", formatter_class=SaneFormatter)
  subparsers = parser.add_subparsers(dest="subcommand")

  common = argparse.ArgumentParser(add_help=False, formatter_class=SaneFormatter)
  common.add_argument("ngsfiles", type=str, nargs="+", help="Next-generation sequencing files to process (BAM or FASTQ).")
  common.add_argument("-o", "--outfile", type=str, help="Write to filename rather than standard out.", default="stdout")
  common.add_argument("--delimiter", default="<tab>", help="Delimiter for the outfile.")
  common.add_argument("--debug", default=False, action="store_true", help="Enable DEBUG log level.")
  common.add_argument("-v", "--verbose", default=False, action="store_true", help="Enable INFO log level.")
  common.add_argument('--version', action='version', version='%(prog)s 1.0.0')

  readlen = subparsers.add_parser("readlen", parents=[common], formatter_class=SaneFormatter)
  readlen.add_argument("-c", "--majority-vote-cutoff", type=float, help="To call a majority readlen, the maximum read length must have at least `majority-vote-cutoff`%% reads in support.", default=0.7)
  readlen.add_argument("-n", "--n-samples", type=int, help="How many reads to sample.", default=1000000)

  instrument = subparsers.add_parser("instrument", parents=[common], formatter_class=SaneFormatter)
  instrument.add_argument("-n", "--n-samples", type=int, help="How many reads to sample.", default=10000)

  strandedness = subparsers.add_parser("strandedness", parents=[common], formatter_class=SaneFormatter)
  strandedness.add_argument("-g", "--gene-model", help="Gene model as a GFF/GTF file.", required=True)
  strandedness.add_argument("--max-tries", type=int, default=3, help="When inconclusive, the test will repeat until this many tries have been reached.")
  strandedness.add_argument("--max-iterations-per-try", type=int, default=1000, help="At most, search this many times for genes that satisfy our search criteria.")
  strandedness.add_argument("-m", "--minimum-reads-per-gene", type=int, help="Filter any genes that don't have at least `m` reads.", default=10)
  strandedness.add_argument("-n", "--n-genes", type=int, help="How many genes to sample.", default=100)
  strandedness.add_argument("-q", "--min-mapq", type=int, help="Minimum MAPQ to consider for reads.", default=30)
  protein_coding_parser = strandedness.add_mutually_exclusive_group(required=False)
  protein_coding_parser.add_argument('--only-protein-coding-genes', dest='only_protein_coding_genes', action='store_true', help="Only consider protein coding genes")
  protein_coding_parser.add_argument('--no-only-protein-coding-genes', dest='only_protein_coding_genes', action='store_false')
  split_by_rg_parser = strandedness.add_mutually_exclusive_group(required=False)
  split_by_rg_parser.add_argument('--split-by-rg', dest='split_by_rg', action='store_true', help="Contain one entry per read group.")
  split_by_rg_parser.add_argument('--no-split-by-rg', dest='split_by_rg', action='store_false')
  strandedness.set_defaults(only_protein_coding_genes=True, split_by_rg=False)
  
  args = parser.parse_args()
  if not args.subcommand:
    parser.print_help()
    sys.exit(1)

  return args


def setup_logging(log_level=logging.INFO):
    """Set up the logging.
    
    Forked from MIT code here: https://github.com/MisterWil/abodepy.
    """
    logging.basicConfig(level=log_level)
    fmt = ("%(asctime)s %(levelname)s (%(threadName)s) "
           "[%(name)s] %(message)s")
    colorfmt = "%(log_color)s{}%(reset)s".format(fmt)
    datefmt = '%H:%M:%S'

    try:
        from colorlog import ColoredFormatter
        logging.getLogger().handlers[0].setFormatter(ColoredFormatter(
            colorfmt,
            datefmt=datefmt,
            reset=True,
            log_colors={
                'DEBUG': 'cyan',
                'INFO': 'green',
                'WARNING': 'yellow',
                'ERROR': 'red',
                'CRITICAL': 'red',
            }
        ))
    except ImportError:
        pass

    logger = logging.getLogger('')
    logger.setLevel(log_level)

def process_args(args):
  # setup logging
  log_level = logging.WARN
  
  if args.verbose:
    log_level = logging.INFO

  if args.debug:
    log_level = logging.DEBUG

  setup_logging(log_level)

  # set output file
  if args.outfile == "stdout":
    args.outfile = sys.stdout
  else:
    args.outfile = open(args.outfile, "w")

  # set delimiter
  if args.delimiter == "<tab>":
    args.delimiter = "\t"


def run():
  args = get_args()
  process_args(args)

  if args.subcommand == "readlen":
    readlen.main(args.ngsfiles,
                 outfile=args.outfile,
                 delimiter=args.delimiter,
                 n_samples=args.n_samples,
                 majority_vote_cutoff=args.majority_vote_cutoff)
  if args.subcommand == "instrument":
    instrument.main(args.ngsfiles, 
                    outfile=args.outfile,
                    delimiter=args.delimiter,
                    n_samples=args.n_samples)
  if args.subcommand == "strandedness":
    strandedness.main(args.ngsfiles, args.gene_model, 
                      outfile=args.outfile,
                      delimiter=args.delimiter,
                      n_genes=args.n_genes,
                      minimum_reads_per_gene=args.minimum_reads_per_gene,
                      only_protein_coding_genes=args.only_protein_coding_genes,
                      min_mapq=args.min_mapq,
                      split_by_rg=args.split_by_rg,
                      max_iterations_per_try=args.max_iterations_per_try)
